#!/usr/bin/env python

import sys
import os.path
import numpy as np
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from jmarkov.dtmc import dtmc


P = np.array([[0.2, 0.7, 0, 0.1], [0.3, 0.4, 0.2,0.1], [0.4, 0, 0.5,0.1],[0.25,0.25,0.25,0.25]])

mc = dtmc(P)
print(mc.transition_matrix)

ss = mc.steady_state()
print(ss)
probas=mc.transient_probabilities(3,np.array([0.05,0.9,0,0.05]))
print(probas)


#Let's check the ergodicity of the chain by calculating its period and checking if it is irreducible:
print('Period of the chain:')
print(mc.period())
print('Is the chain irreducible?')
print(mc.is_irreducible())
print('Is the chain ergodic?')
print(mc.is_ergodic())

#Let's check the oergodicity of some other examples of aperiodic chains
print('\n')
print('Let\'s check the ergodicity of some other examples of aperiodic chains')
print('\n')
P = np.array([[0.1, 0.4, 0.5, 0.0, 0.0, 0.0],
              [0.0, 0.2, 0.3, 0.5, 0.0, 0.0],
              [0.0, 0.0, 0.1, 0.4, 0.5, 0.0],
              [0.0, 0.0, 0.0, 0.2, 0.3, 0.5],
              [0.5, 0.0, 0.0, 0.0, 0.2, 0.3],
              [0.3, 0.5, 0.0, 0.0, 0.0, 0.2]])
P = dtmc(P)
print('Period of the chain:')
print(P.period())
print('Is the chain irreducible?')
print(P.is_irreducible())
print('Is the chain ergodic?')
print(P.is_ergodic())

print('\n')

P = np.array([[0, 1/3, 2/3],
              [0,0,1],
              [1,0,0]])
P = dtmc(P)
print('Period of the chain:')
print(P.period())
print('Is the chain irreducible?')
print(P.is_irreducible())
print('Is the chain ergodic?')
print(P.is_ergodic())

#Now, let's check the ergodicity of some non-aperiodic chains:
print('\n')
print('Now, let\'s check the ergodicity of some non-aperiodic chains:')
print('\n')

#Chain of period 2:
P = np.array([[0, 0, 0.8, 0.2],
              [0, 0, 0.4, 0.6],
              [0.7, 0.3, 0, 0],
              [0.2, 0.8, 0, 0]])
P = dtmc(P)
print('Period of the chain:')
print(P.period())
print('Is the chain irreducible?')
print(P.is_irreducible())
print('Is the chain ergodic?')
print(P.is_ergodic())

#Chain of period 3:
print('\n')
P = np.array([[0, 0, 0.5, 0.25, 0.25, 0, 0],
              [0, 0, 1/3, 0, 2/3, 0, 0],
              [0, 0, 0, 0, 0, 1/3, 2/3],
              [0, 0, 0, 0, 0, 1/2, 1/2],
              [0, 0, 0, 0, 0, 3/4, 1/4],
              [0.5, 0.5, 0, 0, 0, 0, 0],
              [0.25, 0.75, 0, 0, 0, 0, 0]])
P = dtmc(P)
print('Period of the chain:')
print(P.period())
print('Is the chain irreducible?')
print(P.is_irreducible())
print('Is the chain ergodic?')
print(P.is_ergodic())


#Chain of period 4:
print('\n')
P = np.array([[0, 1, 0, 0],
              [0, 0, 1, 0],
              [0, 0, 0, 1],
              [1, 0, 0, 0]])
P = dtmc(P)
print('Period of the chain:')
print(P.period())
print('Is the chain irreducible?')
print(P.is_irreducible())
print('Is the chain ergodic?')
print(P.is_ergodic())

print('\n')


#Lets try the ocupation time function:
M=mc.occupation_time(10)
print("ocupation_matrix:")
print(M)

F=dtmc(np.array([[0.8,0.2,0],[0,0.2,0.8],[0.8,0.2,0]]))
print("Lets try first passage time in dtcm with this matrix: ")
print(np.array([[0.8,0.2,0],[0,0.2,0.8],[0.8,0.2,0]]))
print("This is the result: ")
print(F.first_passage_time(0))
O=dtmc(np.array([[0.9,0.1],[0.2,0.8]]))
print("Lets try another first passage time in dtcm with this matrix: ")
print(np.array([[0.9,0.1],[0.2,0.8]]))
print("This is the result: ")
print(O.first_passage_time(1))

print("lets try an error creating a dtcm")
error_mc=dtmc(np.array([[0.5,0.5],[0.7,0.7]]))


import unittest
import sys
import os.path
import numpy as np
from numpy.testing import assert_allclose, assert_almost_equal
# context for the tests
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from jmarkov.queue.mmk import mmk


class TestSteadyState(unittest.TestCase):
    def test_mmk_mean_number_system(self):
        lda = 1 
        mu = 2
        k =3
        q = mmk(k,lda,mu) 
        assert_almost_equal(q.mean_number_entities(), 0.5030303030, 
                        err_msg="should be 0.5030303030]")

    def test_mmk_mean_number_queue(self):
        lda = 1 
        mu = 2
        k =3
        q = mmk(k,lda,mu) 
        assert_almost_equal(q.mean_number_entities_queue(), 0.0030303030, 
                        err_msg="should be 0.0030303030]")
    
    def test_mmk_mean_number_service(self):
        lda = 1 
        mu = 2
        k =3
        q = mmk(k,lda,mu) 
        assert_almost_equal(q.mean_number_entities_service(), 0.5, 
                        err_msg="should be 0.5]")

    def test_mmk_mean_time_system(self):
        lda = 1 
        mu = 2
        k =3
        q = mmk(k,lda,mu) 
        assert_almost_equal(q.mean_time_system(), 0.5030303030, 
                        err_msg="should be 0.5030303030]")

    def test_mmk_mean_time_queue(self):
        lda = 1 
        mu = 2
        k =3
        q = mmk(k,lda,mu) 
        assert_almost_equal(q.mean_time_queue(), 0.0030303030, 
                        err_msg="should be 0.0030303030]")
    
    def test_mmk_mean_time_service(self):
        lda = 1 
        mu = 2
        k =3
        q = mmk(k,lda,mu) 
        assert_almost_equal(q.mean_time_service(), 0.5, 
                        err_msg="should be 0.5]")


if __name__ == '__main__':
    unittest.main()


import numpy as np
from dtmc import dtmc


P = np.array([[0.2, 0.8, 0], [0.3, 0.4, 0.3], [0.4, 0, 0.6]])
mc = dtmc(P)
#print(mc.transition_matrix)
ss = mc.steady_state()
probas=mc.transient_probabilities(1,np.array([0.3,0.4,0.3]))
print(probas)

P_neg = np.array([[0.2, -0.8, 0], [0.3, 0.4, 0.3], [0.4, 0, 0.6]])
dtmc(P_neg)

P_sumOver1 = np.array([[0.2, 0.8, 0], [0.309, 0.4, 0.3], [0.4, 0, 0.6]])
dtmc(P_sumOver1)





import numpy as np
from jmarkov.ctbd import ctbd

class mmk():
    """
    Implements an M/M/k queue and computes steady state metrics 
    
    The M/M/k queue has exponential interarrival times, with rate arr_rate,
    exponential service times, with ser_rate,
    and k servers in parallel.

    The class builds a birth-death chain that models this queue and uses 
    the steady state probability distribution of the chain to compute
    measures of performance such as mean number of entities in the system,
    in queue, in service, and the mean time in the system, in queue, and in service.
    """
    # number of servers
    k:int

    # arrival rate
    arr_rate:np.float64

    # service rate
    ser_rate:np.float64

    # initializer 
    def __init__(self, k:int, arr_rate:np.float64, ser_rate:np.float64):
        """
        Creates an M/M/k queue with k servers, arr_rate arrival rate and ser_rate service rate
        """
        self.k=k
        self.arr_rate = arr_rate
        self.ser_rate = ser_rate

    def mean_number_entities(self)-> np.float64:
        """
        Computes the mean number of entities in the system in steady state
        
        A birth-death chain is built and its stattionary probability distribution is used 
        to compute the mean number of entities in the system in steady state
        """
        if self.is_stable():
            birth = np.ones(self.k)*self.arr_rate
            death = np.arange(1,self.k+1)*self.ser_rate
            bd = ctbd(birth, death)
            n = 100
            probs = bd.steady_state(n)
            mean_num = 0 
            for i in range(n):
                mean_num += probs[i]*i

            return mean_num
        else:
            print('Unstable queue')
            return 0


    def mean_number_entities_queue(self)-> np.float64:
        """
        Computes the mean number of entities in queue in steady state
        
        A birth-death chain is built and its stattionary probability distribution is used 
        to compute the mean number of entities in queue in steady state
        """
        if self.is_stable():
            birth = np.ones(self.k)*self.arr_rate
            death = np.arange(1,self.k+1)*self.ser_rate
            bd = ctbd(birth, death)
            n = 100
            probs = bd.steady_state(n)
            mean_num = 0 
            for i in range(self.k, n):
                mean_num += probs[i]*(i-self.k)

            return mean_num
        else:
            print('Unstable queue')
            return 0

    def mean_number_entities_service(self)-> np.float64:
        """
        Computes the mean number of entities in service in steady state
        
        A birth-death chain is built and its stattionary probability distribution is used 
        to compute the mean number of entities in the system in steady state
        """
        if self.is_stable():
            return self.arr_rate/self.ser_rate
        else:
            print('Unstable queue')
            return 0
        
    def mean_time_system(self)-> np.float64:
        """
        Computes the mean time in the system in steady state
        
        A birth-death chain is built and its stationary probability distribution is used 
        to compute the mean number of entities in the system in steady state, which is 
        then used with Little's Law to obtain the mean time in the system in steady state
        """
        if self.is_stable():
            L = self.mean_number_entities()
            return L/self.arr_rate
        else:
            print('Unstable queue')
            return 0
    
    def mean_time_queue(self)-> np.float64:
        """
        Computes the mean time in the queue in steady state
        
        A birth-death chain is built and its stationary probability distribution is used 
        to compute the mean number of entities in the queue in steady state, which is 
        then used with Little's Law to obtain the mean time in the system in steady state
        """
        if self.is_stable():
            Lq = self.mean_number_entities_queue()
            return Lq/self.arr_rate
        else:
            print('Unstable queue')
            return 0
    
    def mean_time_service(self)-> np.float64:
        """
        Computes the mean time in service
        
        A simple relation is used to obtain the mean service time
        """
        if self.is_stable():
            return 1/self.ser_rate
        else:
            print('Unstable queue')
            return 0
    


    def is_stable(self)-> bool:
        """
        Returns True if the queue is stable, False otherwise
        
        This queue is stable if the arrival rate is smaller than the 
        maximum service rate
        """
        return self.arr_rate < self.k*self.ser_rate 
        

    



import numpy as np
from scipy import linalg
from scipy import sparse
from .markov_chain import markov_chain

class finite_ctbd():
    """
    Implements a finite continuous-time birth-death (CTBD) Markov chain 
    
    The finite CTBD chain is defined by its birth and death rates, which themselves
    define the number of states, states, and the generator matrix. 
    The representation in this class relies on the birth and death rates between
    states 0 to K.  
    The class provides methods to compute stationary metrics, as well as to check the 
    chain properties (ergodicity). 
    """

    # birth rates states in float array form
    birth_rates:np.array = np.array([1])

    # death rates states in float array form
    death_rates:np.array = np.array([1])
    
    # number of states in string array form
    n_states:int=1
    
    # empty initializer 
    # BD chain with a single state
    def __init__(self):
        self.birth_rates = np.array([0.5])
        self.death_rates = np.array([1])

    # initializer with birth and death rates up to state K
    def __init__(self,birth_rates:np.array,death_rates:np.array):
        """
        Creates a finite  continuous-time Birth-Death Markov chain from its birth and death rates
        """
        if not self._check_birth_death_rates(birth_rates,death_rates): 
            raise ValueError("The birth and death rate are not positive or do not have the same dimension")
        self.birth_rates = birth_rates
        self.death_rates = death_rates
        self.n_states=len(self.birth_rates)+1

    def _check_birth_death_rates(self, birth_rates:np.ndarray, death_rates:np.ndarray):
        """
        Checks that all birth and death rates are positive 
         
        Checks that arrays of birth and death rates have the same nonzero size
        """
        if (birth_rates > 0).all() and (death_rates > 0).all() and len(birth_rates) == len(death_rates):
            return True
        else:
            return False


    def steady_state(self) -> np.ndarray:
        """
        Computes the steady state distribution of the continuous-time BD Markov chain

        Computes the steady state probability distribution by solving the balance equations
        to first find pi0 and then the remaining entries of the distribution. 

        Returns the stationary probability distribution in array form
        """
        if self.is_ergodic():
            # Compute cumulative products of birth and death rates
            birth_prod = np.cumprod(self.birth_rates)
            death_prod = np.cumprod(self.death_rates)
            # Compute C factors as division of cumulative products
            c_factors = np.divide(birth_prod,death_prod)
            # Compute pi_0: probability that the chain is empty in steady state
            pi0 = 1/(1+np.sum(c_factors))
            # Compute remaining entries of pi up to state K
            pi = pi0*c_factors
            pi = np.append(np.array(pi0), pi)
            return pi
        else:    
            print("Birth-death Markov chain is not ergodic")
        return 0
    

    def is_irreducible(self):
        """
        Checks if the given continous-time BD Markov Chain is irreducible.
        
        This method determines if the birth-death continous-time Markov Chain is irreducible 
        by checking if, starting in any state, it is possible to reach any other state 
        in a sequence of transitions.

        Returns a boolean
        """
        return self._check_birth_death_rates(self.birth_rates, self.death_rates)

    def is_ergodic(self):
        """
        Checks if the given continous-time BD Markov Chain is ergodic.
        
        Given that a finite continous-time Markov Chain is ergodic if it is irreducible, this
        method checks if it is irreducible to determine if the provided chain is ergodic or not.
        """
        # We check if the chain is irreducible (sufficient condition as the chain is finite)
        return self.is_irreducible()



import unittest
import sys
import os.path
import numpy as np
from numpy.testing import assert_allclose
# context for the tests
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from jmarkov.dtmc import dtmc


class TestSteadyState(unittest.TestCase):
    def test_dtmc_steady_state(self):
        P = np.array([[0.2, 0.8, 0], [0.3, 0.4, 0.3], [0.4, 0, 0.6]])
        mc = dtmc(P)
        assert_allclose(mc.steady_state(), [0.3, 0.4, 0.3],err_msg="should be [0.3, 0.4, 0.3]")

class TestPeriod(unittest.TestCase):
    def test_period_1(self):
        P = np.array([[0.1, 0.4, 0.5, 0.0, 0.0, 0.0],
                      [0.0, 0.2, 0.3, 0.5, 0.0, 0.0],
                      [0.0, 0.0, 0.1, 0.4, 0.5, 0.0],
                      [0.0, 0.0, 0.0, 0.2, 0.3, 0.5],
                      [0.5, 0.0, 0.0, 0.0, 0.2, 0.3],
                      [0.3, 0.5, 0.0, 0.0, 0.0, 0.2]])
        mc = dtmc(P)
        self.assertEqual(mc.period(), 1)
        
    def test_period_2(self):
        P = np.array([[0, 0, 0.8, 0.2],
                      [0, 0, 0.4, 0.6],
                      [0.7, 0.3, 0, 0],
                      [0.2, 0.8, 0, 0]])
        mc = dtmc(P)
        self.assertEqual(mc.period(), 2)
    
    def test_period_3(self):
        P = np.array([[0, 0, 0.5, 0.25, 0.25, 0, 0],
                      [0, 0, 1/3, 0, 2/3, 0, 0],
                      [0, 0, 0, 0, 0, 1/3, 2/3],
                      [0, 0, 0, 0, 0, 1/2, 1/2],
                      [0, 0, 0, 0, 0, 3/4, 1/4],
                      [0.5, 0.5, 0, 0, 0, 0, 0],
                      [0.25, 0.75, 0, 0, 0, 0, 0]])
        mc = dtmc(P)
        self.assertEqual(mc.period(), 3)
    
    def test_period_4(self):
        P = np.array([[0, 1, 0, 0],
                      [0, 0, 1, 0],
                      [0, 0, 0, 1],
                      [1, 0, 0, 0]])
        mc = dtmc(P)
        self.assertEqual(mc.period(), 4)

class TestIsIrreducible(unittest.TestCase):
    def test_dtmc_isirreducible(self):
        P = np.array([[0, 1/3, 2/3],
                      [0,0,1],
                      [1,0,0]])
        mc = dtmc(P)
        self.assertTrue(mc.is_irreducible())

class TestIsErgodic(unittest.TestCase):
    def test_dtmc_isergodic_true(self):
        P = np.array([[0, 1/3, 2/3],
                      [0,0,1],
                      [1,0,0]])
        mc = dtmc(P)
        self.assertTrue(mc.is_ergodic())

    def test_dtmc_isergodic_false(self):
        P = np.array([[0, 1, 0, 0],
                      [0, 0, 1, 0],
                      [0, 0, 0, 1],
                      [1, 0, 0, 0]])
        mc = dtmc(P)
        self.assertFalse(mc.is_ergodic())

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

import sys
import os.path
import numpy as np
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from jmarkov.queue.mmk import mmk

lda = 1 
mu = 2
k =3

q = mmk(k,lda,mu) 
L = q.mean_number_entities()

print("Mean number of entities in system:")
print(L)

Lq = q.mean_number_entities_queue()
print("Mean number of entities in queue:")
print(Lq)



import sys
import os.path
import numpy as np
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..','..','..')))
from jmarkov.mdp.dtmdp import dtmdp

# Variable de estado:
# X_{t}: Grado de deterioro de la máquina en la época t

# Espacio de estados
estadosMaquina = np.array(['Perfectas Condiciones','Defecto','Avería Total'])
# Acciones
accionesMaquina = np.array(['Reparar','Sustituir','Nada'])
# Retornos Inmediatos
retornosMaquina = []
for estado_actual in range(0,len(estadosMaquina)):
    listaAux = []
    for a in range(0,len(accionesMaquina)):
        # Si decide reparar
        if a == 0:
            listaAux.append(50)
        # Si decide sustituir
        elif a == 1:
            listaAux.append(80)
        # Si decide no hacer nada y está en defecto
        elif a == 2 and estado_actual == 1:
            listaAux.append(10)
        # Si decide no hacer nada y está en avería total
        elif a == 2 and estado_actual == 2:
            listaAux.append(20)
        else:
            listaAux.append(0)
    retornosMaquina.append(listaAux)
retornosMaquina = np.array(retornosMaquina)

# Matrices de transición
matricesMaquina = {}
# Para a == 0
matrizMaquina = []
for estado_actual in range(0,len(estadosMaquina)):
    listaAux = []
    for estado_futuro in range(0,len(estadosMaquina)):
        # Si está en perfectas condiciones
        if estado_actual == 0 and estado_futuro == 0:
            listaAux.append(1)
        # Si está en defecto y queda en perfectas condiciones
        elif estado_actual == 1 and estado_futuro == 0:
            listaAux.append(0.8)
        # Si está en defecto y permanece en defecto
        elif estado_actual == 1 and estado_futuro == 1:
            listaAux.append(0.2)
        # Si está en avería total y queda en perfectas condiciones
        elif estado_actual == 2 and estado_futuro == 0:
            listaAux.append(0.3)
        # Si está en avería total y permanece en avería total
        elif estado_actual == 2 and estado_futuro == 2:
            listaAux.append(0.7)
        else:
            listaAux.append(0)
    matrizMaquina.append(listaAux)
    matricesMaquina[accionesMaquina[0]] = np.array(matrizMaquina)
    
# Para a == 1
matrizMaquina = []
for estado_actual in range(0,len(estadosMaquina)):
    listaAux = []
    for estado_futuro in range(0,len(estadosMaquina)):
        # Si está en perfectas condiciones
        if estado_futuro == 0:
            listaAux.append(1)
        else:
            listaAux.append(0)
    matrizMaquina.append(listaAux)
    matricesMaquina[accionesMaquina[1]] = np.array(matrizMaquina)    

# Para a == 2
matrizMaquina = []
for estado_actual in range(0,len(estadosMaquina)):
    listaAux = []
    for estado_futuro in range(0,len(estadosMaquina)):
        # Si está en perfectas condiciones y permanece en perfectas condiciones
        if estado_actual == 0 and estado_futuro == 0:
            listaAux.append(0.9)
        # Si está en perfectas condiciones y presenta defecto
        elif estado_actual == 0 and estado_futuro == 1:
            listaAux.append(0.09)
        # Si está en perfectas condiciones y presenta avería total
        elif estado_actual == 0 and estado_futuro == 2:
            listaAux.append(0.01)
        # Si está en defecto y permanece en defecto
        elif estado_actual == 1 and estado_futuro == 1:
            listaAux.append(0.55)
        # Si está en defecto y presenta avería total
        elif estado_actual == 1 and estado_futuro == 2:
            listaAux.append(0.45)
        # Si está en avería total y permanece en avería total
        elif estado_actual == 2 and estado_futuro == 2:
            listaAux.append(1)
        else:
            listaAux.append(0)
    matrizMaquina.append(listaAux)
    matricesMaquina[accionesMaquina[2]] = np.array(matrizMaquina)
    
# Creo el problema como un mdp
mdpMaquina = dtmdp(estadosMaquina, accionesMaquina, matricesMaquina, retornosMaquina, 0.8)
print(mdpMaquina.solve(0, minimize = True, method = "value_iteration"))
print(mdpMaquina.solve(0, minimize = True, method="policy_iteration"))



import sys
import os.path
import numpy as np
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..','..','..')))
from jmarkov.mdp.dtmdp import dtmdp

# Parámetros:
lambdaP = 10
lambdaB = 20
lambdaE = 30
lambdaF = 40
lambdaTotal = lambdaP + lambdaB + lambdaE + lambdaF
lambdaPB = lambdaP + lambdaB

w = [20, 15, 5, 2.5]

# Variable de estado:
# X_{t}: tipo de encargo que recibe un rappitendero en la época t

# Espacio de Estados
estadosRappi = np.array(['Premium', 'Bundle', 'Estándar','RappiFavor'])

# Espacio de Acciones
accionesRappi = np.array(['Aceptar', 'Rechazar'])

# Retornos Inmediatos
retornosRappi = []
for i in range(0,len(estadosRappi)):
    listaAux = []
    for a in range(0,len(accionesRappi)):
        if accionesRappi[a] == 'Rechazar':
            listaAux.append(0)
        else:
            listaAux.append(w[i])
    retornosRappi.append(listaAux)
retornosRappi = np.array(retornosRappi)

# Probabilidades de Transición

# Creo el diccionario con matrices vacías
matricesRappi = {}
for a in range(0,len(accionesRappi)):
    matricesRappi[accionesRappi[a]] = np.zeros((len(estadosRappi),len(estadosRappi)))
# Recorro sobre las acciones
for a in range(0,len(accionesRappi)):
    # Recorro sobre el estado actual
    for estadoActual in range(0,len(estadosRappi)):
        # Recorro sobre el estado futuro 
        for estadoFuturo in range(0, len(estadosRappi)):
            if accionesRappi[a] == 'Aceptar':
                if estadosRappi[estadoFuturo] == 'Premium':
                    matricesRappi[accionesRappi[a]][estadoActual][estadoFuturo] = lambdaP/lambdaTotal
                elif estadosRappi[estadoFuturo] == 'Bundle':
                    matricesRappi[accionesRappi[a]][estadoActual][estadoFuturo] = lambdaB/lambdaTotal
                elif estadosRappi[estadoFuturo] == 'Estándar':
                    matricesRappi[accionesRappi[a]][estadoActual][estadoFuturo] = lambdaE/lambdaTotal
                elif estadosRappi[estadoFuturo] == 'RappiFavor':
                    matricesRappi[accionesRappi[a]][estadoActual][estadoFuturo] = lambdaF/lambdaTotal
                else:
                    matricesRappi[accionesRappi[a]][estadoActual][estadoFuturo] = 0
            elif accionesRappi[a] == 'Rechazar':
                if estadosRappi[estadoFuturo] == 'Premium':
                    matricesRappi[accionesRappi[a]][estadoActual][estadoFuturo] = lambdaP/lambdaPB
                elif estadosRappi[estadoFuturo] == 'Bundle':
                    matricesRappi[accionesRappi[a]][estadoActual][estadoFuturo] = lambdaB/lambdaPB
                else:
                    matricesRappi[accionesRappi[a]][estadoActual][estadoFuturo] = 0

# Creo el problema como un mdp
mdpRappi = dtmdp(estadosRappi, accionesRappi, matricesRappi, retornosRappi, 0.8)
print(mdpRappi.solve(0, minimize = False, method = "value_iteration"))
print(mdpRappi.solve(0, minimize = False, method="policy_iteration"))


from abc import ABC, abstractmethod

class markov_chain(ABC):

    @property
    @abstractmethod
    def n_states(self):
        # number of states
        pass

    @property
    @abstractmethod
    def states(self):
        # markov chain states
        pass

    @abstractmethod
    def steady_state(self):
        # computes the steady state distribution
        pass

    @abstractmethod
    def first_passage_time(self, target:str):
        # computes the expected first passage time to target state
        pass

    @abstractmethod
    def occupation_time(self, nsteps:int):
        # computes the expected occupation time matrix in nsteps steps
        pass

    @abstractmethod
    def is_ergodic(self):
        # determines id the chain is ergodic or not
        pass




import sys
import os.path
import numpy as np
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..','..','..')))
from jmarkov.sdp.dtsdp import dtsdp


#Vector de épocas
E = np.array([i for i in range(1,4)])
# Variables
estados = np.array(["Excelente","Bueno","Promedio","Malo"]) # Estado de la máquina al inicio de la semana
# Decisiones
A = np.array(["Reemplazar","No Reemplazar"])
# Retornos Inmediatos
R = np.zeros((len(E), len(estados), len(A)))
for t in range(len(E)): 
    for s_index,i in enumerate(estados):
        for posA,a in enumerate(A):
            if(i=="Excelente" and a=="Reemplazar"):
                R[t,s_index,posA]=-1000000
            elif(i=="Excelente" and a=="No Reemplazar"):
                R[t,s_index,posA]=100
            elif(i=="Bueno" and a=="Reemplazar"):
                R[t,s_index,posA]=-100
            elif(i=="Bueno" and a=="No Reemplazar"):
                R[t,s_index,posA]=80
            elif(i=="Promedio" and a=="Reemplazar"):
                R[t,s_index,posA]=-100
            elif(i=="Promedio" and a=="No Reemplazar"):
                R[t,s_index,posA]=50
            elif(i=="Malo" and a=="Reemplazar"):
                R[t,s_index,posA]=-100
            elif(i=="Malo" and a=="No Reemplazar"):
                R[t,s_index,posA]=10

# Matrices de transición
probs = {t:np.zeros((len(A), len(estados), len(estados))) for t in E}

matNoReemplazar = np.array([[0.7,0.3,0,0],
                          [0,0.7,0.3,0],
                          [0,0,0.7,0.3],
                          [0,0,0,1]])

matReemplazar = np.array([[1,0,0,0],
                          [0.7,0.3,0,0],
                          [0.7,0.3,0,0],
                          [0.7,0.3,0,0]])
probs = {}
for t in E:  # Iterar sobre cada época
    decisiones_dict = {}
    for posA, a in enumerate(A):
        if a == "Reemplazar":
            decisiones_dict[a] = matReemplazar
        elif a == "No Reemplazar":
            decisiones_dict[a] = matNoReemplazar
    probs[t] = decisiones_dict

sdpMaquinas = dtsdp(E,estados,A,probs,R,0.9)
print(sdpMaquinas.solve(minimize = False))


import sys
import os.path
import numpy as np
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..','..','..')))
from jmarkov.sdp.dtsdp import dtsdp

#Vector de épocas
E = np.array([i for i in range(1,31)])

# Variables
Sx = np.array([i for i in range(90000,-9000,-9000)])   
Sy = np.array([i for i in range(0,33,3)])   
Sw = np.array([i for i in range(3000,18000,3000)]) 
estados = [(i, j, k) for i in Sx for j in Sy for k in Sw if (3000*j<=90000-i)] # Variable conjunta

# Decisiones
A = np.array([3,0,-3])

# Retornos Inmediatos
R = np.zeros((len(E),len(estados), len(A)))

#(250, 300, 350, 400, 450)
dividendos = {}
dividendos[3000] =250 
dividendos[6000] =300 
dividendos[9000] =350 
dividendos[12000] =400 
dividendos[15000] =450 

def retornos(i,j,k,a):
    retorno = -1e8
    if j>=-1*a and i>=a*k:
        retorno = j*dividendos[k] - a*k
    return retorno
for t in E:
    for posS, s in enumerate(estados):
        i = s[0]
        j = s[1]
        k = s[2]
        for posA, a in enumerate(A):
            R[t-1,posS,posA] = retornos(i,j,k,a)

# Matrices de probabilidades de transición
p1 = np.array([
    [0.333, 0.555, 0.111, 0, 0],
    [0.222, 0.444, 0.277, 0.055, 0],
    [0.181, 0.181, 0.272, 0.363, 0],
    [0, 0.222, 0.222, 0.222, 0.333],
    [0, 0, 0, 0.75, 0.25]
])
p2 = np.array([
    [0.2, 0.4, 0.4, 0, 0],
    [0.083, 0.416, 0.333, 0.166, 0],
    [0.125, 0.1875, 0.375, 0.25, 0.0625],
    [0, 0.153, 0.230, 0.384, 0.230],
    [0, 0, 0.2, 0.4, 0.4]
])
p3 = np.array([
    [0.25, 0.25, 0.5, 0, 0],
    [0.166, 0.1667, 0.333, 0.333, 0],
    [0.083, 0.25, 0.166, 0.333, 0.166],
    [0, 0.0625, 0.25, 0.25, 0.4375],
    [0, 0, 0.153, 0.461, 0.384]
])

sz_indices = {value: idx for idx, value in enumerate(Sw)}

# Probabilidades de transición
probs = {}
for t in E:  
    decisiones_dict = {}
    if t <= 10:
        matriz = p1
    elif t >10 and t <= 20:
        matriz = p2
    else: 
        matriz = p3
    for posA, a in enumerate(A):
        matriz_transicion = np.zeros((len(estados), len(estados)))
        for fila, estado_inicial in enumerate(estados):
            i = estado_inicial[0]
            j = estado_inicial[1]
            k = estado_inicial[2]
            for columna, estado_futuro in enumerate(estados):
                i_ = estado_futuro[0]
                j_ = estado_futuro[1]
                k_ = estado_futuro[2]
                
                if(j>=-1*a and i>=a*k and i_ == i - max(0,a)*k and j_ == j+a):
                    k_idx = sz_indices[k]
                    k_idx_ = sz_indices[k_]
                    matriz_transicion[fila, columna] = matriz[k_idx][k_idx_]
        decisiones_dict[a] = matriz_transicion
    probs[t] = decisiones_dict

sdpBitcoin = dtsdp(E,estados,A,probs,R,0.99)
print(sdpBitcoin.solve(minimize = False))


#!/usr/bin/env python

import sys
import os.path
import numpy as np
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from jmarkov.ctbd import ctbd

birth = np.array([1])
death = np.array([2])
bd = ctbd(birth, death)

print("Check if the BD chain is ergodic:")
print(bd.is_ergodic())

print("Compute the steady state distribution of the BD chain:")
ss = bd.steady_state()
print(ss)

n=4
print(f"Compute {n} entries of the steady state distribution of the BD chain:")
ssn = bd.steady_state(n)
print(ssn)

ss10 = ss.copy()
for i in range(100):
    ss10 = np.append(ss10, np.array(ss10[-1]*birth[-1]/death[-1]))
#print(ss10)
#print(sum(ss10))





import numpy as np
from typing import Dict


class dtsdp():
    """
    Implements an Stochastic Decision Process (SDP)  
    
    The process is defined by its time horizon, number of states, states, number of actions, actions,
    the immediate returns of implementing each action and a transition matrix for each action. 
    The class provides methods to solve SDPs and to compute the expected value of the optimal policy.     
    """
    # time steps in array form
    periods:np.array=np.array([1])
    # number of time steps
    n_periods:int=1
    # number of states in string array form
    n_states:int=1
    # states in string array form
    states:np.array = np.array([1])
    # number of actions
    n_actions:int=1
    # actions in array form
    actions:np.array = np.array([1])
    # transition matrix as dict of numpy arrays
    transition_matrices:Dict[str, np.array]
    # immediate return matrix as 3d numpy array 
    immediate_returns:np.array = np.array([[[1]]])
    # discount factor as int
    discount_factor:float = 0.9
    # initializer with a transition matrix, immediate returns and discount factor
    def __init__(self,periods:np.array,states:np.array, actions: np.array, transition_matrices:Dict, immediate_returns: np.array,discount_factor:int):
        """
        Creates a markov decision process from its transition matrices, immediate returns and discount factor
        """
        if not self._check_transition_matrices(transition_matrices):#Lets check if transition matrix is logical (i.e the rows sum 1)
            raise ValueError("the rows of transition matrices do not sum 1 or have non positive values")
        if not self._check_immediate_returns(immediate_returns,states,periods,actions): # Lets check if immediate returns are consistent
            raise ValueError("the dimensions of the immediate returns are not coherent")
        if not self._check_discount_factor(discount_factor): # Lets check if discount factor is logical
            raise ValueError("discount factor should be a number between 0 and 1")
        if not self._check_time_period(periods):
            raise ValueError("the time period should be greater than 0")
        self.periods = periods
        self.n_periods = len(periods)
        self.n_actions = len(actions)
        self.n_states=len(states)
        self.transition_matrices = transition_matrices
        self.immediate_returns = immediate_returns
        self.discount_factor = discount_factor
        self.actions = actions
        self.states = states
    def _check_transition_matrices(self, M:Dict):
        """
        Checks that all matrices are stochastic
         
        Checks that all row sums are equal to one and all elements are non negative
        """
        # determines if the transition matrices are logical
        # i.e: if the sum of the rows equals to 1 (is an stochastic matrix)
        # and if all of the elements are positive
        for epoch, matrices in M.items():
            for action, probabilities in matrices.items():
                for state in range(probabilities.shape[0]):
                    row_sum = probabilities[state,].sum()
                    if not np.allclose(row_sum,1,atol=1e-6) and not np.all((probabilities[state,] >= 0) & (probabilities[state,] <= 1)):
                        return False
                    else:
                        return True
    def _check_immediate_returns(self,R:np.array,states:np.array,periods:np.array,actions:np.array):  
        """
        Checks that the immediate returns are valid and dimensionally-coherent
         
        Checks that immediate return array has dimensions length of epochs x length of states x length of actions
        """  
        # checks if the dimensions of immediate returns are coherent
        # expected: len(E)xlen(S) x len(actions)
        if R.shape == (len(periods),len(states),len(actions)):
            return True

    def _check_discount_factor(self,beta:int):
        """
        Checks that the discount factor is valid
         
        Checks that discount factor is a number equal to or greater than 0 and  less than 1
        """  
        # checks if the discount factor is a number between 0 and 1 
        if beta >= 0 and beta <= 1:
            return True
        else: 
            return False

    def _check_time_period(self,periods:int):
        """"
        Checks that the time period is valid

        Checks that the time period is a number greater than 0
        """     
        # checks if the time period is greater than 0
        if len(periods) > 0:
            return True
        else:
            return False
        
    def solve(self, minimize = False):
        """
        Solves SDP's with backward iteration

        Returns the expected value of following the optimal policy at each state and the optimal policy for each state
        """
        E = self.periods
        S = self.states
        A = self.actions
        M = self.transition_matrices
        R = self.immediate_returns
        beta = self.discount_factor
        if minimize == True:
            R = -R
            best_init = 100000
        else:
            best_init = -100000

        # creates an array to save the value of Bellman's equation
        # for each state, for each time period
        Ft_optimo = np.zeros((len(S),len(E)))
        # creates an array to save the optimal decision 
        # for each state, for each time period
        Mat_Dec_optimo = np.empty((len(S),len(E)), dtype=str)

        # to begin backward iteration, it is required to initialize
        # the last time period
        f = R
        for s_index,i in enumerate(S):
            Ft_optimo[s_index,-1] = max(f[-1,s_index])
            dec = int(np.where(Ft_optimo[s_index,-1] == max(f[-1,s_index]))[0][0])
            Mat_Dec_optimo[s_index,-1] = A[dec]
        # start backward iteration
        # iterate through time steps (from second to last, to first)
        for t in range(len(E)-2,-1,-1):
            # iterate through states
            for s_index,i in enumerate(S):
                # initialize the value of bellman's equation
                best_value = best_init
                # initialize the optimal decision
                a_optima = len(A)-1
                # iterate through decisions
                for posA, a in enumerate(A):
                    expected_value = R[t,s_index,posA] + beta*np.sum(M[t+1][a][s_index,:]*Ft_optimo[:,t+1])
                    # check if it promises improvement
                    if expected_value>best_value:
                        # if it improves, update the values
                        best_value = expected_value
                        a_optima = posA
                # update de matrix of bellman equation and optimal decisions
                Ft_optimo[s_index,t] = best_value
                Mat_Dec_optimo[s_index,t] = A[a_optima]  
        if minimize == True:
            Ft_optimo = -Ft_optimo   
        return(Ft_optimo,Mat_Dec_optimo)     


import unittest
import sys
import os.path
import numpy as np
from numpy.testing import assert_allclose
# context for the tests
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from jmarkov.mdp.dtmdp import dtmdp

class TestSolver(unittest.TestCase):
    def test_V_value_iteration(self):
        # number of states:
        N = 2
        # states:
        states = np.array([i for i in range(0,N)])
        # actions
        actions = np.array([a for a in range(0,N)]) 
        # immediate returns:
        immediate_returns = np.array([[3, 1], [2, 3]])
        # discount factor:
        discount_factor = 0.8
        # transition matrices
        transition_matrices = {}
        transition_matrices["0"] = np.array([[1/2, 1/2],[1/3, 2/3]])
        transition_matrices["1"] = np.array([[1/4, 3/4],[2/3, 1/3]])

        mdp = dtmdp(states, actions, transition_matrices, immediate_returns, discount_factor)
        result = mdp.solve(0, minimize = True)[0]
        assert_allclose(result, ([7.81249996, 8.74999996]),err_msg="should be ([7.81249996, 8.74999996])")
    def test_policy_value_iteration(self):
        # number of states:
        N = 2
        # states:
        states = np.array([i for i in range(0,N)])
        # actions
        actions = np.array([a for a in range(0,N)]) 
        # immediate returns:
        immediate_returns = np.array([[3, 1], [2, 3]])
        # discount factor:
        discount_factor = 0.8
        # transition matrices
        transition_matrices = {}
        transition_matrices["0"] = np.array([[1/2, 1/2],[1/3, 2/3]])
        transition_matrices["1"] = np.array([[1/4, 3/4],[2/3, 1/3]])

        mdp = dtmdp(states, actions, transition_matrices, immediate_returns, discount_factor)
        result = mdp.solve(0, minimize = True)[1]
        self.assertEqual(result, {0: 1, 1: 0})
    def test_V_policy_iteration(self):
        # number of states:
        N = 2
        # states:
        states = np.array([i for i in range(0,N)])
        # actions
        actions = np.array([a for a in range(0,N)]) 
        # immediate returns:
        immediate_returns = np.array([[3, 1], [2, 3]])
        # discount factor:
        discount_factor = 0.8
        # transition matrices
        transition_matrices = {}
        transition_matrices["0"] = np.array([[1/2, 1/2],[1/3, 2/3]])
        transition_matrices["1"] = np.array([[1/4, 3/4],[2/3, 1/3]])

        mdp = dtmdp(states, actions, transition_matrices, immediate_returns, discount_factor)
        result = mdp.solve(0, minimize = True)[0]
        assert_allclose(result, ([7.81249996, 8.74999996]),err_msg="should be ([7.81249996, 8.74999996])")
    def test_policy_policy_iteration(self):
        # number of states:
        N = 2
        # states:
        states = np.array([i for i in range(0,N)])
        # actions
        actions = np.array([a for a in range(0,N)]) 
        # immediate returns:
        immediate_returns = np.array([[3, 1], [2, 3]])
        # discount factor:
        discount_factor = 0.8
        # transition matrices
        transition_matrices = {}
        transition_matrices["0"] = np.array([[1/2, 1/2],[1/3, 2/3]])
        transition_matrices["1"] = np.array([[1/4, 3/4],[2/3, 1/3]])

        mdp = dtmdp(states, actions, transition_matrices, immediate_returns, discount_factor)
        result = mdp.solve(0, minimize = True)[1]
        self.assertEqual(result, {0: 1, 1: 0})
        
if __name__ == '__main__':
    unittest.main()



import numpy as np
from scipy import linalg
from scipy import sparse
from .markov_chain import markov_chain

class ctbd():
    """
    Implements an infinite continuous-time birth-death (CTBD) Markov chain 
    
    The CTBD chain is defined by its birth and death rates, which themselves
    define the number of states, states, and the generator matrix. 
    The representation in this class relies on the birth and death rates between
    states 0 to K, and it is assumed that from state K onwards the birth and death rates
    have the same values.  
    The class provides methods to compute both
    stationary and transient metrics, as well as to check the 
    chain properties (ergodicity). 
    """

    # birth rates states in float array form
    birth_rates:np.array = np.array([1])

    # death rates states in float array form
    death_rates:np.array = np.array([1])
 
    # empty initializer 
    # BD chain with a single state
    def __init__(self):
        self.birth_rates = np.array([0.5])
        self.death_rates = np.array([1])

    # initializer with birth and death rates up to state K
    def __init__(self,birth_rates:np.array,death_rates:np.array):
        """
        Creates a continuous-time Birth-Death Markov chain from its birth and death rates
        """
        if not self._check_birth_death_rates(birth_rates,death_rates): 
            raise ValueError("The birth and death rate are not positive or do not have the same dimension")
        self.birth_rates = birth_rates
        self.death_rates = death_rates

    def _check_birth_death_rates(self, birth_rates:np.ndarray, death_rates:np.ndarray):
        """
        Checks that all birth and death rates are positive 
         
        Checks that arrays of birth and death rates have the same nonzero size
        """
        if (birth_rates > 0).all() and (death_rates > 0).all() and len(birth_rates) == len(death_rates):
            return True
        else:
            return False


    def steady_state(self, n:int=0) -> np.ndarray:
        """
        Computes the steady state distribution of the continuous-time BD Markov chain

        Computes the steady state probability distribution by solving the balance equations
        to first find pi0 and then the remaining entries of the distribution. 

        n is an optional parameter to specify the length of the steady state distribution 
        vector. It should be at least as large as the size of the birth and death rate arrays. 
        If not specified, the returned vector is equal to the length of the birth and death arrays. 

        Returns the stationary probability distribution in array form
        """
        if self.is_ergodic():
            # Compute cumulative products of birth and death rates
            birth_prod = np.cumprod(self.birth_rates)
            death_prod = np.cumprod(self.death_rates)
            # Compute C factors as division of cumulative products
            c_factors = np.divide(birth_prod,death_prod)
            # Compute pi_0: probability that the chain is empty in steady state
            pi0 = 1/(1+np.sum(c_factors[:-1])+c_factors[-1]*self.death_rates[-1]/(self.death_rates[-1]-self.birth_rates[-1]))
            # Compute remaining entries of pi up to length of birth/death rates arrays
            pi = pi0*c_factors
            pi = np.append(np.array(pi0), pi)
            # Compute remaining entries pi up to n    
            final_factor = self.birth_rates[-1]/self.death_rates[-1]
            if n > len(self.birth_rates):
                for _ in range(n-len(self.birth_rates)-1):
                    pi = np.append(pi, pi[-1]*final_factor)
            return pi
        else:    
            print("Birth-death Markov chain is not ergodic")
        return 0
    

    def is_irreducible(self):
        """
        Checks if the given continous-time BD Markov Chain is irreducible.
        
        This method determines if the birth-death continous-time Markov Chain is irreducible 
        by checking if, starting in any state, it is possible to reach any other state 
        in a sequence of transitions.

        Returns a boolean
        """
        return self._check_birth_death_rates(self.birth_rates, self.death_rates)

    def is_ergodic(self):
        """
        Checks if the given continous-time BD Markov Chain is ergodic.
        
        Given that this chain is infinite, we first check if is irreducible 
        and next check if the last birth rate is smaller than the last 
        death rate, a condition that ensures ergodicity.
        """
        # We check if the chain is irreducible
        if self.is_irreducible()==True:
            return self.death_rates[-1] > self.birth_rates[-1]
        else:
            return False



import unittest
import sys
import os.path
import numpy as np
from numpy.testing import assert_allclose, assert_almost_equal
# context for the tests
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from jmarkov.queue.mmkn import mmkn


class TestSteadyState(unittest.TestCase):
    # tests for M/M/1/5
    def test_mm15_mean_number_system(self):
        lda = 1 
        mu = 2
        k =1
        n = 5
        q = mmkn(k,lda,mu,n) 
        assert_almost_equal(q.mean_number_entities(), 0.904761905, 
                        err_msg="should be 0.904761905]")

    def test_mm15_mean_number_queue(self):
        lda = 1 
        mu = 2
        k = 1
        n = 5
        q = mmkn(k,lda,mu,n) 
        assert_almost_equal(q.mean_number_entities_queue(), 0.412698413, 
                        err_msg="should be 0.412698413]")
    
    def test_mm15_mean_number_service(self):
        lda = 1 
        mu = 2
        k = 1
        n = 5
        q = mmkn(k,lda,mu,n) 
        assert_almost_equal(q.mean_number_entities_service(), 0.492063492, 
                        err_msg="should be 0.492063492]")

    def test_mm15_mean_time_system(self):
        lda = 1 
        mu = 2
        k = 1
        n = 5
        q = mmkn(k,lda,mu,n) 
        assert_almost_equal(q.mean_time_system(), 0.919354839, 
                        err_msg="should be 0.919354839]")

    def test_mm15_mean_time_queue(self):
        lda = 1 
        mu = 2
        k = 1
        n = 5
        q = mmkn(k,lda,mu,n) 
        assert_almost_equal(q.mean_time_queue(), 0.419354839, 
                        err_msg="should be 0.419354839]")
    
    def test_mm15_mean_time_service(self):
        lda = 1 
        mu = 2
        k = 1
        n = 5
        q = mmkn(k,lda,mu,n) 
        assert_almost_equal(q.mean_time_service(), 0.5, 
                        err_msg="should be 0.5]")

    # tests for M/M/2/5
    def test_mm25_mean_number_system(self):
        lda = 1 
        mu = 2
        k = 2
        n = 5
        q = mmkn(k,lda,mu,n) 
        assert_almost_equal(q.mean_number_entities(), 0.531066822978, 
                        err_msg="should be 0.531066822978]")

    def test_mm25_mean_number_queue(self):
        lda = 1 
        mu = 2
        k = 2
        n = 5
        q = mmkn(k,lda,mu,n) 
        assert_almost_equal(q.mean_number_entities_queue(), 0.031652989449, 
                        err_msg="should be 0.031652989449]")
    
    def test_mm25_mean_number_service(self):
        lda = 1 
        mu = 2
        k = 2
        n = 5
        q = mmkn(k,lda,mu,n) 
        assert_almost_equal(q.mean_number_entities_service(), 0.499413833529, 
                        err_msg="should be 0.499413833529]")

    def test_mm25_mean_time_system(self):
        lda = 1 
        mu = 2
        k = 2
        n = 5
        q = mmkn(k,lda,mu,n) 
        assert_almost_equal(q.mean_time_system(), 0.531690140845, 
                        err_msg="should be 0.531690140845]")

    def test_mm25_mean_time_queue(self):
        lda = 1 
        mu = 2
        k = 2
        n = 5
        q = mmkn(k,lda,mu,n) 
        assert_almost_equal(q.mean_time_queue(), 0.031690140845, 
                        err_msg="should be 0.031690140845]")
    
    def test_mm25_mean_time_service(self):
        lda = 1 
        mu = 2
        k = 2
        n = 5
        q = mmkn(k,lda,mu,n) 
        assert_almost_equal(q.mean_time_service(), 0.5, 
                        err_msg="should be 0.5]")

if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python

import sys
import os.path
import numpy as np
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from jmarkov.queue.mmkn import mmkn

# M/M/1/5

lda = 1 
mu = 2
k = 1
n = 5

q = mmkn(k,lda,mu,n) 
L = q.mean_number_entities()
print("Mean number of entities in system:")
print(L)

Lq = q.mean_number_entities_queue()
print("Mean number of entities in queue:")
print(Lq)

Ls = q.mean_number_entities_service()
print("Mean number of entities in service:")
print(Ls)

W = q.mean_time_system()
print("Mean time in system:")
print(W)

Wq = q.mean_time_queue()
print("Mean time in queue:")
print(Wq)

Ws = q.mean_time_service()
print("Mean time in service:")
print(Ws)

# M/M/2/5

lda = 1 
mu = 2
k = 2
n = 5

q = mmkn(k,lda,mu,n) 
L = q.mean_number_entities()
print("Mean number of entities in system:")
print(L)

Lq = q.mean_number_entities_queue()
print("Mean number of entities in queue:")
print(Lq)

Ls = q.mean_number_entities_service()
print("Mean number of entities in service:")
print(Ls)

W = q.mean_time_system()
print("Mean time in system:")
print(W)

Wq = q.mean_time_queue()
print("Mean time in queue:")
print(Wq)

Ws = q.mean_time_service()
print("Mean time in service:")
print(Ws)



import numpy as np
from scipy import linalg
from scipy import sparse
from .markov_chain import markov_chain

class ctmc(markov_chain):
    """
    Implements a finite continuous-time Markov chain (CTMC) 
    
    The chain is defined by its number of states, states, and a 
    generator matrix. The class provides methods to compute both
    stationary and transient metrics, as well as to check the 
    chain properties (ergodicity). 
    """

    # number of states in string array form
    n_states:int=1

    # states in string array form
    states:np.array = np.array([1])

    # generator matrix as 2d numpy array
    generator:np.array
    
    # empty initializer 
    # chain with a single state
    def __init__(self):
        self.n_states = 1
        self.generator = np.array([0])

    # initializer with only the number of states
    def __init__(self,n:int):
        self.n_states = n
        self.generator = np.zeros((n,n), dtype=float)

    # initializer with a generator matrix
    def __init__(self,generator:np.array):
        """
        Creates a continuous-time Markov chain from its generator matrix
        """
        if not self._check_generator_matrix(generator):#Lets check if transition matrix is logical (i.e the rows sum 0)
            raise ValueError("the rows of generator matrix do not sum 0 or the diagonal has non negative values")
        self.n_states=generator.shape[0]
        self.generator = generator


    def _check_generator_matrix(self,M:np.ndarray):
        """
        Checks that a matrix is a Markov chain generator
         
        Checks that all row sums are equal to zero, diagonal entries are non-positive, and 
        non-diagonal entries are non-negative
        """
        #
        vector = np.isclose(np.sum(M, axis = 1),0,1e-5) == True
        # Check if a given transition matrix has all diagonal elements non positive
        if vector.all() and np.all(np.diag(M)<0):
            return True
        else:
            return False


    def steady_state(self) -> np.ndarray:
        """
        Computes the steady state distribution of the continuous-time Markov chain

        Computes the steady state probability distribution by replacing one of the 
        matrix equations with a normalizing equation that ensures the result is a 
        probability distribution. 

        Returns the stationary probability distribution in array form
        """

        # 
        if self.generator.any():
            # sets A to be Q and replace first column with ones for the normalizing equation
            A = self.generator.copy()
            A[:, 0] =  1
            # sets b as the right hand vector with a 1 for the normalizing equation
            b = np.zeros(self.n_states, dtype=float)
            b[0] = 1
            # solves transposed system as pi is a row vector
            res = linalg.solve(A,b,transposed=True)
            return res
        else:    
            print("Empty generator matrix")
        return 0
    

    def transient_probabilities(self,t:float,alpha:np.ndarray) -> np.ndarray:
        """
        Computes the transient distribution at time t with initial state alpha

        Computes alpha*exp(Q*t)*ones to obtain the probability distribution
        at time t given the initial probability distribution alpha
        """
        shape=alpha.shape#lets get the shape of alpha vector
        if shape[0]!=self.n_states:
            raise ValueError("The dimensions of alpha vector are incorrect. It must be a vector 1xn_states")
        P=linalg.expm(self.generator*t)#exponentiate the diferential generator following the method of Mohy et al (https://eprints.maths.manchester.ac.uk/1300/1/alhi09a.pdf)
        probs=alpha@P
        return probs


    def first_passage_time(self, target:str):
        """
        Computes the expected first passage time to a target state from a start state.

        This method calculates the expected number of steps required for the Markov chain to reach
        the specified target state from the start state by creating a sub-matrix of the generator matrix with the target removed.

        Returns the expected steps to reach the target state from any start state (except target) in array form        
        """
        #The transition matrix and the number of states are brought
        e=self.n_states 
        q=self.generator.copy()
        
        # Matrix of ones with n-1 rows is created
        u=np.full([e-1,1],1)

        #The column and row corresponding to the target state are eliminated from the transition matrix 
        m=np.delete(q,target, axis=0)
        m=np.delete(m,target, axis=1)


        t=np.matmul(np.linalg.inv(-m),u)
        return t

    def occupation_time(self,T,Epsilon=0.00001):
        n=self.n_states
        m=self.generator.copy()
        def P_from_R(n,m):            
            # Inicializamos el valor máximo con el primer elemento de la diagonal
            Vector_ri=np.diagonal(m)   
            #Hallamos la r como el máximo de las r_i
            r=max(-Vector_ri)
            return r 
        def creacion_P(n,m):
            r= P_from_R(n,m)
            #Hallamos la matriz P^        
            h=np.zeros((n,n))
            for i in range(n):
                for j in range(n):
                    if i==j:
                        h[i][j]=1+(m[i][j]/r)
                    else:
                        h[i][j]=m[i][j]/r
            return h 
        r=P_from_R(n,m)
        P=creacion_P(n,m)
        i=np.identity(n)
        A=P
        k=0
        yek=np.exp(-r*T)
        ygk=1-yek
        suma=ygk
        B=ygk*i
        while suma/r < T-Epsilon:
            k=k+1
            yek=yek*(r*T)/k
            ygk=ygk-yek
            B=B+ygk*A
            A=A@P
            suma=suma+ygk
        return (B/r)
    

    def is_irreducible(self):
        """
        Checks if the given continous-time Markov Chain is irreducible.
        
        This method determines if the continous-time Markov Chain is irreducible by checking if, starting in
        any state, it is possible to reach any other state in a sequence of transitions.

        Returns a boolean
        """
        # the finite case:

        # We bring the generator matrix and replace all the diagonal elements for 0
        Q = np.copy(self.generator)
        np.fill_diagonal(Q,0)
        # We check if all the components in the chain are strongly connected
        if sparse.csgraph.connected_components(Q, directed=True,connection='strong',return_labels=False)==1:
            return True
        else:
            return False

    def is_ergodic(self):
        """
        Checks if a given continous-time Markov Chain is ergodic.
        
        Given that a finite continous-time Markov Chain is ergodic if it is irreducible, this
        method checks if it is irreducible to determine if the provided chain is ergodic or not.
        """
        # We check if the chain is irreducible
        if self.is_irreducible()==True:
            return True
        else:
            return False



import unittest
import sys
import os.path
import numpy as np
from numpy.testing import assert_allclose
# context for the tests
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from jmarkov.sdp.dtsdp import dtsdp
class TestSolver(unittest.TestCase):
    def test_value_solver(self):
            # number of epochs:
            epochs = np.array([i for i in range(1,4)])
            # states:
            states = np.array(["Excelente","Bueno","Promedio","Malo"])
            # actions
            actions = np.array(["Reemplazar","No Reemplazar"]) 
            # immediate returns:
            immediate_returns = np.zeros((len(epochs), len(states), len(actions)))
            for t in range(len(epochs)): 
                for s_index,i in enumerate(states):
                    for posA,a in enumerate(actions):
                        if(i=="Excelente" and a=="Reemplazar"):
                            immediate_returns[t,s_index,posA]=-1000000
                        elif(i=="Excelente" and a=="No Reemplazar"):
                            immediate_returns[t,s_index,posA]=100
                        elif(i=="Bueno" and a=="Reemplazar"):
                            immediate_returns[t,s_index,posA]=-100
                        elif(i=="Bueno" and a=="No Reemplazar"):
                            immediate_returns[t,s_index,posA]=80
                        elif(i=="Promedio" and a=="Reemplazar"):
                            immediate_returns[t,s_index,posA]=-100
                        elif(i=="Promedio" and a=="No Reemplazar"):
                            immediate_returns[t,s_index,posA]=50
                        elif(i=="Malo" and a=="Reemplazar"):
                            immediate_returns[t,s_index,posA]=-100
                        elif(i=="Malo" and a=="No Reemplazar"):
                            immediate_returns[t,s_index,posA]=10
            # discount factor:
            discount_factor = 1
            # transition matrices
            transition_matrices = {t:np.zeros((len(actions), len(states), len(states))) for t in epochs}

            matNoReemplazar = np.array([[0.7,0.3,0,0],
                                    [0,0.7,0.3,0],
                                    [0,0,0.6,0.4],
                                    [0,0,0,1]])

            matReemplazar = np.array([[0,0,0,0],
                                    [0.7,0.3,0,0],
                                    [0.7,0.3,0,0],
                                    [0.7,0.3,0,0]])
            transition_matrices = {}
            for t in epochs:  # Iterar sobre cada época
                decisiones_dict = {}
                for posA, a in enumerate(actions):
                    if a == "Reemplazar":
                        decisiones_dict[a] = matReemplazar
                    elif a == "No Reemplazar":
                        decisiones_dict[a] = matNoReemplazar
                transition_matrices[t] = decisiones_dict
            sdp = dtsdp(epochs, states, actions, transition_matrices, immediate_returns, discount_factor)
            result = sdp.solve(minimize = False)[0]
            assert_allclose(result, ([281.1, 194. , 100. ],[210.9, 151. ,  80. ],[108.4,  84. ,  50. ],[ 81.1,  20. ,  10. ]),err_msg="should be ([281.1, 194. , 100. ],[210.9, 151. ,  80. ],[108.4,  84. ,  50. ],[ 81.1,  20. ,  10. ])")
    def test_policy_solver(self):
            # number of epochs:
            epochs = np.array([i for i in range(1,4)])
            # states:
            states = np.array(["Excelente","Bueno","Promedio","Malo"])
            # actions
            actions = np.array(["Reemplazar","No Reemplazar"]) 
            # immediate returns:
            immediate_returns = np.zeros((len(epochs), len(states), len(actions)))
            for t in range(len(epochs)): 
                for s_index,i in enumerate(states):
                    for posA,a in enumerate(actions):
                        if(i=="Excelente" and a=="Reemplazar"):
                            immediate_returns[t,s_index,posA]=-1000000
                        elif(i=="Excelente" and a=="No Reemplazar"):
                            immediate_returns[t,s_index,posA]=100
                        elif(i=="Bueno" and a=="Reemplazar"):
                            immediate_returns[t,s_index,posA]=-100
                        elif(i=="Bueno" and a=="No Reemplazar"):
                            immediate_returns[t,s_index,posA]=80
                        elif(i=="Promedio" and a=="Reemplazar"):
                            immediate_returns[t,s_index,posA]=-100
                        elif(i=="Promedio" and a=="No Reemplazar"):
                            immediate_returns[t,s_index,posA]=50
                        elif(i=="Malo" and a=="Reemplazar"):
                            immediate_returns[t,s_index,posA]=-100
                        elif(i=="Malo" and a=="No Reemplazar"):
                            immediate_returns[t,s_index,posA]=10
            # discount factor:
            discount_factor = 1
            # transition matrices
            transition_matrices = {t:np.zeros((len(actions), len(states), len(states))) for t in epochs}

            matNoReemplazar = np.array([[0.7,0.3,0,0],
                                    [0,0.7,0.3,0],
                                    [0,0,0.6,0.4],
                                    [0,0,0,1]])

            matReemplazar = np.array([[0,0,0,0],
                                    [0.7,0.3,0,0],
                                    [0.7,0.3,0,0],
                                    [0.7,0.3,0,0]])
            transition_matrices = {}
            for t in epochs:  # Iterar sobre cada época
                decisiones_dict = {}
                for posA, a in enumerate(actions):
                    if a == "Reemplazar":
                        decisiones_dict[a] = matReemplazar
                    elif a == "No Reemplazar":
                        decisiones_dict[a] = matNoReemplazar
                transition_matrices[t] = decisiones_dict
            sdp = dtsdp(epochs, states, actions, transition_matrices, immediate_returns, discount_factor)
            result = sdp.solve(minimize = False)[1]
            self.assertTrue(np.array_equal(result, ([['N', 'N', 'R'],['N', 'N', 'R'],['N', 'N', 'R'],['R', 'N', 'R']])))
if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python

import sys
import os.path
import numpy as np
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from jmarkov.ctmc import ctmc

# Creating the Q of a drive Thru with one queue space, mic and window
Q = np.array([[-50,50, 0, 0, 0, 0, 0, 0],
[0, -80, 50, 30, 0, 0, 0, 0],
[0, 0, -30, 0, 30, 0, 0, 0],
[40, 0, 0, -90, 50, 0, 0, 0],
[0, 40, 0, 0, -120, 50, 30, 0],
[0, 0, 40, 0, 0, -70, 0, 30],
[0, 0, 0, 40, 0, 0, -90, 50],
[0, 0, 0, 0, 40, 0, 0, -40]])

mc = ctmc(Q)
print(mc.generator)

ss = mc.steady_state()
print(ss)
print("lets try the transient method:")
a=mc.transient_probabilities(1,np.array([1/8,1/8,1/8,1/8,1/8,1/8,1/8,1/8]))
print(a)



import sys
import os.path
import numpy as np
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..','..','..')))
from jmarkov.ctmc import ctmc

lambdaLlegada = 0.5
mu = 1/4
theta = 1/45
omega = 1/10

# Espacio de estados
Sx = [i for i in range(0,7)] # [0, 1, 2, 3, 4, 5, 6, 7]
Sy = ["R","W"] 

estados = []
for i in Sx:
    for j in Sy:
        estado = ",".join((str(i),str(j)))
        estados.append(estado)
estados = np.array(estados)

matriz = []
for estado_actual in estados:
    i = int(estado_actual[0])
    j = estado_actual[-1]
    listaAux = []
    for estado_futuro in estados:
        i_ = int(estado_futuro[0])
        j_ = estado_futuro[-1]
        if i_==i+1 and j_==j and i<6:
            listaAux.append(lambdaLlegada)
        elif i_==i-1 and j_==j and i>0 and j=="W":
            listaAux.append(mu)
        elif i_==i and j_=="R" and j=="W":
            listaAux.append(theta)
        elif i_==i and j_=="W" and j=="R":
            listaAux.append(omega)
        else:
            listaAux.append(0)
    matriz.append(listaAux)

for estado_actual in range(0, len(estados)):
    for estado_futuro in range(0, len(estados)):
        if estado_actual == estado_futuro:
            matriz[estado_actual][estado_futuro]=-sum(matriz[estado_actual])

matriz = np.array(matriz)

elefante = ctmc(matriz)



import numpy as np
from typing import Dict
import sys
import os.path
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..','..')))
from jmarkov.dtmc import dtmc

class dtmdp():
    """
    Implements a Markov decision process (MDP)  
    
    The process is defined by its number of states, states, number of actions, actions,
    the immediate returns of implementing each action and a transition matrix for each action. 
    The class provides methods to solve MDPs and to compute the expected value of the optimal policy. 
    """
    # number of states in string array form
    n_states:int=1
    # states in string array form
    states:np.array = np.array([1])
    # number of actions
    n_actions:int=1
    # actions in array form
    actions:np.array = np.array([1])
    # transition matrix as dict of numpy arrays
    transition_matrices:Dict[str, np.array]
    # immediate return matrix as 2d numpy array 
    immediate_returns:np.array = np.array([1])   
    # discount factor as int
    discount_factor:int = 0.8 
    # initializer with a transition matrix, immediate returns and discount factor
    def __init__(self,states:np.array, actions: np.array, transition_matrices:Dict, immediate_returns: np.array,discount_factor:int):
        """
        Creates a markov decision process from its transition matrices, immediate returns and discount factor
        """
        if not self._check_transition_matrices(transition_matrices):#Lets check if transition matrix is logical (i.e the rows sum 1)
            raise ValueError("the rows of transition matrices do not sum 1 or have non positive values")
        if not self._check_immediate_returns(immediate_returns,transition_matrices): # Lets check if immediate returns are consistent
            raise ValueError("the dimensions of the immediate returns are not coherent")
        if not self._check_discount_factor(discount_factor): # Lets check if discount factor is logical
            raise ValueError("discount factor should be a number between 0 and 1")
        self.n_actions = len(transition_matrices)
        self.n_states=len(transition_matrices[next(iter(transition_matrices))])
        self.transition_matrices = transition_matrices
        self.immediate_returns = immediate_returns
        self.discount_factor = discount_factor
        self.actions = actions
        self.states = states

    def _check_transition_matrices(self, M:Dict):
        """
        Checks that all matrices are stochastic
         
        Checks that all row sums are equal to one and all elements are non negative
        """
        # determines if the transition matrices are logical
        # i.e: if the sum of the rows equals to 1 (is an stochastic matrix)
        # and if all of the elements are positive
        if all(np.all((np.array(value) >= 0) & (np.array(value) <= 1)) for value in M.values()) and np.all(np.isclose(np.sum(list(M.values()), axis=2), 1, atol=1e-5)):
            return True
        else:
            return False
        
    def _check_immediate_returns(self,R:np.array, M:Dict):  
        """
        Checks that the immediate returns are valid and dimensionally-coherent
         
        Checks that immediate return array has dimensions length of states x length of actions
        """  
        # checks if the dimensions of immediate returns are coherent
        # expected: len(S) x len(actions)
        if R.shape == (len(list(M.values())[0]),len(M)):
            return True

    def _check_discount_factor(self,beta:int):
        """
        Checks that the discount factor is valid
         
        Checks that discount factor is a number equal to or greater than 0 and  less than 1
        """  
        # checks if the discount factor is a number between 0 and 1 
        if beta >= 0 and beta < 1:
            return True
        else: 
            return False
    
    def solve(self, tolerance, minimize = False, method = "value_iteration"):
        """
        Solves MDP's with defined method

        Returns the expected value of following the optimal policy at each state and the optimal policy for each state
        """
        S = self.states
        A = self.actions
        M = self.transition_matrices
        R = self.immediate_returns
        beta = self.discount_factor

        # if the sense of the search is minimize, change the values of the immediate returns
        # so that we maximize minimums
        if minimize == True:
            R = -1*R
        
        if method == "value_iteration":
            result = value_iteration(S, A, M, R, beta, tolerance)
            V = result[0]
            optimal_policy = result[1]
        elif method == "policy_iteration":
            result = policy_iteration(S, A, M, R, beta, tolerance)
            V = result[0]
            optimal_policy = result[1]

        if minimize == True:
            V = -1 * V
        # return the optimal policy
        return V, optimal_policy
    
    def policy_transition_matrix(self, policy):
        """
        Calculates the expected value of a policy according 

        Returns the transition matrix for following the optimal policy
        """
        # gets the transition matrix for the optimal policy
        S = self.states
        M = self.transition_matrices
        optimal_transition_matrix = []
        for i in range(0, len(S)):
            optimal_transition_matrix.append(M[policy[S[i]]][i])
        optimal_transition_matrix = np.array(optimal_transition_matrix)
        return optimal_transition_matrix
        
    def expected_policy_value(self, value_function, policy):
        """
        Calculates the expected value of a policy according to the steady state probability of being in each state

        Returns the expected value of following the optimal policy
        """
        # gets the transition matrix for the optimal policy
        optimal_transition_matrix = self.policy_transition_matrix(policy)
        # creates the dtmc object
        mc = dtmc(optimal_transition_matrix)
        expected_value = sum(mc.steady_state()*value_function)
        return expected_value
    
def value_iteration(S, A, M, R, beta, tolerance):
    # initialize value functions
    V = np.zeros(len(S))
    # initialize optimal policy
    optimal_policy = {S[i]: A[0] for i in range(0, len(S))}
    # iterate while there is no improvement
    while True:
        # save values from previous iteration
        oldV = V.copy()
        # iterate through states
        for i in range(0, len(S)):
            # initialize Q -> value-action function
            Q = {}
            # iterate through actions
            for a in range(0,len(A)):
                # evaluate the new value function
                Q[A[a]] = R[i, a] + beta*sum(M[A[a]][i,j]*oldV[j] for j in range(0, len(S)))
                # update the new value function for each state
                V[i] = max(Q.values())
                # update the action for each state
                optimal_policy[S[i]] = max(Q, key = Q.get)

        # if there is no improvement break the cycle
        if np.allclose(oldV, V, tolerance):
            break
    return V, optimal_policy

def policy_improvement(V, S, A, M, R, beta):
    # initialize optimal policy
    optimal_policy = {S[i]: A[0] for i in range(0, len(S))}
    # iterate through states
    for i in range(0, len(S)):
        # initialize Q -> value-action function
        Q = {}
        # iterate through actions
        for a in range(0, len(A)):
            # evaluate the new value function
            Q[A[a]] = R[i,a] + beta*sum(M[A[a]][i,j]*V[j] for j in range(0, len(S)))
        # update the action for each state
        optimal_policy[S[i]] = max(Q, key = Q.get)
    return optimal_policy

def policy_evaluation(policy, S, A, M, R, beta, tolerance):
    # initialize value functions
    V = np.zeros(len(S))
    # iterate while there is improvement
    while True:
        # save old values of value functions
        oldV = V.copy()
        # iterate through states
        for i in range(0,len(S)):
            # save current policy
            a = np.where(A == policy[S[i]])[0][0]
            # evaluate the current policy for each state
            V[i] = R[i,a] + beta*sum(M[A[a]][i,j]*V[j] for j in range(0,len(S)))
        # if there is no improvement break the cycle    
        if np.allclose(oldV, V, tolerance):
            break
    return V

def policy_iteration(S, A, M, R, beta, tolerance):
    # initialize optimal policy
    optimal_policy = {S[i]: A[0] for i in range(0, len(S))}
    # iterate while there is improvement
    while True:
        # save values of current policy
        old_policy = optimal_policy.copy()
        # evaluate current policy
        V = policy_evaluation(optimal_policy, S, A, M, R, beta, tolerance)
        # improve current policy
        optimal_policy = policy_improvement(V, S, A, M, R, beta)
        # if old policy and new policy are the same (there is no improvement) break cycle 
        if all(old_policy[S[i]] == optimal_policy[S[i]] for i in range(0, len(S))):
            break
    return V, optimal_policy



import numpy as np
from jmarkov.finite_ctbd import finite_ctbd

class mmkn():
    """
    Implements an M/M/k/n queue and computes steady state metrics 
    
    The M/M/k/n queue has exponential interarrival times, with rate arr_rate,
    exponential service times, with ser_rate,
    k servers in parallel, and
    n places in total (n-k for waiting).

    The class builds a birth-death chain that models this queue and uses 
    the steady state probability distribution of the chain to compute
    measures of performance such as mean number of entities in the system,
    in queue, in service, and the mean time in the system, in queue, and in service.
    """
    # number of servers
    k:int

    # arrival rate
    arr_rate:np.float64

    # service rate
    ser_rate:np.float64

    # capacity
    n:int

    # effective arrival rate
    eff_arr_rate:np.float64

    # initializer 
    def __init__(self, k:int, arr_rate:np.float64, ser_rate:np.float64, n:int):
        """
        Creates an M/M/k/n queue with k servers, arr_rate arrival rate, ser_rate service rate, and capacity n
        """
        self.k=k
        self.arr_rate = arr_rate
        self.ser_rate = ser_rate
        self.n=n

    def mean_number_entities(self)-> np.float64:
        """
        Computes the mean number of entities in the system in steady state
        
        A birth-death chain is built and its stationary probability distribution is used 
        to compute the mean number of entities in the system in steady state
        """
        probs = self._solve_bd_process()
        mean_num = 0 
        for i in range(self.n+1):
            mean_num += probs[i]*i

        return mean_num


    def mean_number_entities_queue(self)-> np.float64:
        """
        Computes the mean number of entities in queue in steady state
        
        A birth-death chain is built and its stattionary probability distribution is used 
        to compute the mean number of entities in queue in steady state
        """
        probs = self._solve_bd_process()
        mean_num = 0 
        for i in range(self.k, self.n+1):
            mean_num += probs[i]*(i-self.k)

        return mean_num

    def mean_number_entities_service(self)-> np.float64:
        """
        Computes the mean number of entities in service in steady state
        
        A birth-death chain is built and its stattionary probability distribution is used 
        to compute the mean number of entities in the system in steady state
        """
        probs = self._solve_bd_process()
        mean_num = 0 
        for i in range(self.k+1):
            mean_num += probs[i]*i
        for i in range(self.k+1, self.n+1):
            mean_num += probs[i]*self.k
        return mean_num
        
        
    def mean_time_system(self)-> np.float64:
        """
        Computes the mean time in the system in steady state
        
        A birth-death chain is built and its stationary probability distribution is used 
        to compute the mean number of entities in the system in steady state, which is 
        then used with Little's Law to obtain the mean time in the system in steady state
        """
        L = self.mean_number_entities()
        print(self.eff_arr_rate)
        return L/self.eff_arr_rate


    def mean_time_queue(self)-> np.float64:
        """
        Computes the mean time in the queue in steady state
        
        A birth-death chain is built and its stationary probability distribution is used 
        to compute the mean number of entities in the queue in steady state, which is 
        then used with Little's Law to obtain the mean time in the system in steady state
        """
        Lq = self.mean_number_entities_queue()
        return Lq/self.eff_arr_rate


    def mean_time_service(self)-> np.float64:
        """
        Computes the mean time in service
        
        A birth-death chain is built and its stationary probability distribution is used 
        to compute the mean number of entities in the queue in steady state, which is 
        then used with Little's Law to obtain the mean time in the system in steady state
        """
        Ls = self.mean_number_entities_service()
        return Ls/self.eff_arr_rate
   
    def _solve_bd_process(self):
        birth = np.ones(self.n)*self.arr_rate
        death = np.append(np.arange(1,self.k+1)*self.ser_rate, np.ones(self.n-self.k)*self.ser_rate*self.k)
        bd = finite_ctbd(birth, death)
        probs = bd.steady_state()
        self.eff_arr_rate = self.arr_rate*(1-probs[-1])
        return probs



import sys
import os.path
import numpy as np
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..','..','..')))
from jmarkov.mdp.dtmdp import dtmdp


# Parametros:
p = 0.4
q = 0.2
retLleno = 5
retAlto = 4
retMedio = 3
retBajo = 2
W = 10
Z = 4
C = 12

# Variable de estado:
# X_{t}: Nivel del embalse en la época t

# Espacio de estados
#estadosRepresa = [i for i in range(0,4)] # [0, 1, 2, 3] --> 0: Lleno, 1: Alto, 2: Medio, 3: Bajo
estadosRepresa = np.array(['Lleno', 'Alto', 'Medio', 'Bajo'])

# Acciones
#accionesRepresa = np.array([a for a in range(0,2)]) # array([0, 1])--> 0: Abrir, 1:Cerrar
accionesRepresa = np.array(['Abrir','Cerrar'])

# Retornos Inmediatos
retornosRepresa = np.array([[W*retLleno - Z,  -C*(p*q)],
                            [W*retAlto - Z, 0],
                            [W*retMedio - Z, 0],
                            [W*retBajo - Z, 0]])

# Matrices de transición
matricesRepresa = {}
# Para a == 0
matrizRepresa = []
for estado_actual in range(0,len(estadosRepresa)):
    listaAux = []
    for estado_futuro in range(0,len(estadosRepresa)):
        if estado_futuro == 3:
            listaAux.append(1)
        else:
            listaAux.append(0)
    matrizRepresa.append(listaAux)
    matricesRepresa[accionesRepresa[0]] = np.array(matrizRepresa)
    
# Para a == 1
matrizRepresa = []
for estado_actual in range(0,len(estadosRepresa)):
    listaAux = []
    for estado_futuro in range(0,len(estadosRepresa)):
        # Si decido cerrar y estaba lleno
        if estado_futuro == estado_actual and estado_actual>0:
            listaAux.append(1-(p*q))
        # Si decido cerrar y llueve y se aumenta el nivel
        elif estado_futuro == estado_actual-1 and estado_actual>0:
            listaAux.append(p*q)
        # Si decido cerrar y no aumenta el nivel
        elif estado_actual == 0 and estado_futuro == 0:
            listaAux.append(1)
        else:
            listaAux.append(0)
    matrizRepresa.append(listaAux)
    matricesRepresa[accionesRepresa[1]] = np.array(matrizRepresa)

# Creo el problema como un mdp
mdpRepresa = dtmdp(estadosRepresa, accionesRepresa, matricesRepresa, retornosRepresa, 0.8)
print(mdpRepresa.solve(0, method = "value_iteration"))
print(mdpRepresa.solve(0, method="policy_iteration"))


import sys
import os.path
import numpy as np
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..','..','..')))
from jmarkov.ctmc import ctmc

tasaLlegada = 5
tasaSalida = 9
estados = [i for i in range(0,7)]

matriz = []
for i in estados:
    listaAux = []
    for j in estados:
        if j==i+1 and i<6:
            listaAux.append(tasaLlegada)
        elif j==i-1 and i>0:
            listaAux.append(tasaSalida*i)
        else: 
            listaAux.append(0)
    matriz.append(listaAux)
for i in estados:
    for j in estados:
        if i==j:
            matriz[i][j]=-sum(matriz[i])

matriz = np.array(matriz)
rentBike = ctmc(matriz)
  


import sys
import os.path
import numpy as np
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..','..','..')))
from jmarkov.ctmc import ctmc

lam = 3
mu = 2
alpha = 5


# Espacio de estados
Sx = [0, 1] # Vacío (0), Ocupado en la primera fase del servicio (1)
Sy = [0, 1, 2] # Vacío (0), Ocupado en la primera fase del servicio (1), Ocupado en la segunda fase del servicio (2)

estados = []
for i in Sx:
    for j in Sy:
        estado = ",".join((str(i),str(j)))
        estados.append(estado)
estados = np.array(estados)

matriz = []
for estado_actual in estados:
    i = int(estado_actual[0])
    j = int(estado_actual[-1])
    listaAux = []
    for estado_futuro in estados:
        i_ = int(estado_futuro[0])
        j_ = int(estado_futuro[-1])
        
        if i_==1 and j_==j and i==0:
            listaAux.append(0.7*lam)
        elif i_==i and j_==1 and j==0:
            listaAux.append(0.3*lam)
        elif i_==0 and j_==j and i==1:
            listaAux.append(mu)
        elif i_==i and j_==2 and j==1:
            listaAux.append(alpha)
        elif i_==i and j_==0 and j==2:
            listaAux.append(alpha)
        else:
            listaAux.append(0)
    matriz.append(listaAux)
    
for estado_actual in range(0, len(estados)):
    for estado_futuro in range(0, len(estados)):
        if estado_actual == estado_futuro:
            matriz[estado_actual][estado_futuro]=-sum(matriz[estado_actual])

matriz = np.array(matriz)

autosElectricos = ctmc(matriz)


import unittest
import sys
import os.path
import numpy as np
from numpy.testing import assert_allclose
# context for the tests
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from jmarkov.ctbd import ctbd


class TestSteadyState(unittest.TestCase):
    def test_ctdb_steady_state(self):
        birth = np.array([1])
        death = np.array([2])
        bd = ctbd(birth, death)
        assert_allclose(bd.steady_state(), [0.5, 0.25], 
                        err_msg="should be [0.5, 0.25]")

    def test_ctdb_steady_state5(self):
        birth = np.array([1])
        death = np.array([2])
        n=5
        bd = ctbd(birth, death)
        assert_allclose(bd.steady_state(5), [0.5, 0.25, 0.125, 0.0625, 0.03125], 
                        err_msg="should be [0.5, 0.25, 0.125, 0.0625, 0.03125]")

class TestIsErgodic(unittest.TestCase):
    def test_ctdb_isergodic(self):
        birth = np.array([1])
        death = np.array([2])
        bd = ctbd(birth, death)
        self.assertTrue(bd.is_ergodic())

if __name__ == '__main__':
    unittest.main()


import sys
import os.path
import numpy as np
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..','..','..')))
from jmarkov.mdp.dtmdp import dtmdp

# Parametros:
beta = 0.3
alpha = 0.6

# Variables de estado:
# X_{t}: Nivel de carga del robot al inicio de los t-ésimos 15 minutos

# Espacio de estados
estadosRobot = [i for i in range(0,3)] # [0, 1, 2] --> 0: Alto 1: Bajo 2: Descargado

# Acciones
accionesRobot = np.array([a for a in range(0,3)]) # array([0, 1, 2])--> 0: buscar activamente 1: esperar un humano 2: recargar

# Retornos Inmediatos
retornosRobot = []
for i in range(0,len(estadosRobot)):
    listaAux = []
    for a in accionesRobot:
        # Si decide buscar y el nivel de batería es alto
        if a == 0 and i == 0:
            listaAux.append(1)
        # Si decide esperar y el nivel de batería es alto
        elif a == 1 and i == 0:
            listaAux.append(0.5)
        # Si decide recargar y el nivel de batería es alto
        elif a == 2 and i == 0:
            listaAux.append(0)
        # Si decide buscar y el nivel de batería es bajo
        elif a == 0 and i == 1:
            listaAux.append(1)
        # Si decide esperar y el nivel de batería es bajo
        elif a == 1 and i == 1:
            listaAux.append(0.5)
        # Si decide recargar y el nivel de batería es bajo
        elif a == 2 and i == 1:
            listaAux.append(0)
        # Si decide buscar y el nivel de batería es descargado (infactible)
        elif a == 0 and i == 2:
            listaAux.append(-100)
        # Si decide esperar y el nivel de batería es descargado (infactible)
        elif a == 1 and i == 2:
            listaAux.append(-100)
        # Si decide recargar y el nivel de batería es bajo
        elif a == 2 and i == 2:
            listaAux.append(-3)
        else:
            listaAux.append(0)
    retornosRobot.append(listaAux)
retornosRobot = np.array(retornosRobot)

# Matrices de transición
matricesRobot = {}
# Para a == 0
matrizRobot = []
for estado_actual in range(0,len(estadosRobot)):
    listaAux = []
    for estado_futuro in range(0,len(estadosRobot)):
        if estado_actual == 0 and estado_futuro == 0:
            listaAux.append(alpha)
        elif estado_actual == 0 and estado_futuro == 1:
            listaAux.append(1-alpha)
        elif estado_actual == 1 and estado_futuro == 1:
            listaAux.append(1-beta)
        elif estado_actual == 1 and estado_futuro == 2:
            listaAux.append(beta)
        elif estado_actual == 2 and estado_futuro == 2:
            listaAux.append(1)
        else:
            listaAux.append(0)
    matrizRobot.append(listaAux)
    matricesRobot[accionesRobot[0]] = np.array(matrizRobot)
# Para a == 1
matrizRobot = []
for estado_actual in range(0,len(estadosRobot)):
    listaAux = []
    for estado_futuro in range(0,len(estadosRobot)):
        if estado_actual == estado_futuro:
            listaAux.append(1)
        else:
            listaAux.append(0)
    matrizRobot.append(listaAux)
    matricesRobot[accionesRobot[1]] = np.array(matrizRobot)
# Para a == 2
matrizRobot = []
for estado_actual in range(0,len(estadosRobot)):
    listaAux = []
    for estado_futuro in range(0,len(estadosRobot)):
        if estado_futuro == 2:
            listaAux.append(1)
        else:
            listaAux.append(0)
    matrizRobot.append(listaAux)
    matricesRobot[accionesRobot[2]] = np.array(matrizRobot)
    
# Creo el problema como un mdp
mdpRobot = dtmdp(estadosRobot, accionesRobot, matricesRobot, retornosRobot, 0.8)
print(mdpRobot.solve(0, method = "value_iteration"))
print(mdpRobot.solve(0, method="policy_iteration"))